/*
    Tool to create, restore and manage image creation for lab machines

*/

#include <string>
#include <stdlib.h>
#include <cstring>
#include <algorithm>

using namespace std ;
#include <iostream>
#include <fstream>
#include <vector>

class OS {
public:
    string OperatingSystem;
    string Location;
    string Remarks;

};

vector <OS> ReadFile() {
    vector <OS> v1;
    string s1,s2;
    int place;

    ifstream myFile;
    myFile.open( "Configuration" );
    if ( myFile.fail() ) {
        cerr << "Configuration file doesn't exist\n";
        exit(1);
    }
    while(1) {
        getline(myFile,s1, '\n');
        if (myFile.rdstate() & ios::failbit) {
            break;
        }
        //cout << s1;
        // Increase the array, and add another OS
        OS o1;
        place = s1.find_first_of(",");
        //cout << "place = " << place;       
        o1.OperatingSystem = s1.substr(0 ,place) ;
        s1 = s1.substr(place+1,1000);
        //cout << "o1.OperatingSystem = " << o1.OperatingSystem << "\n";
        //  cout << s1;
        place = s1.find_first_of(",");
        //cout << "place = " << place << "\n";       
        o1.Location = s1.substr(0 ,place) ;
        s1 = s1.substr(place+1,1000);
        //cout << "o1.Location = " << o1.Location << "\n";
        o1.Remarks = s1;

        v1.push_back(o1);


    }


    return v1;
}

void PrintOSs(vector <OS> v1)
{
    int i;
    cout << "Number of operating systems is " << v1.size() << "\n";
    for (i = 0; i < v1.size(); i++) {
        cout << "(" << i << ") OperatingSystem = " << v1[i].OperatingSystem << 
            " Location = " << v1[i].Location << 
            " Remarks = " << v1[i].Remarks << "\n";

    }
}

bool ExecuteCommand(string command)
{
    int ret = 0;
    cout << "Executing: " << command << "\n";

#ifndef WIN32
    ret = system (command.c_str());
#endif
    if(ret==0) return true;
    return false;
}

string ImageName(vector <OS> v1, int ImageNumber)
{
    if (ImageNumber >= v1.size()) {
        cerr << "No such image " << ImageNumber << "\n";
    }
    string s1 = "Image";
    s1 += v1[ImageNumber].OperatingSystem + "_"+ v1[ImageNumber].Location + "_"+ v1[ImageNumber].Remarks + ".gzip";
    // I want to replace each / with _ to allow creating legal names
    replace_if(s1.begin(), s1.end(), bind2nd(equal_to<char>(), '/'), '_');
    return s1;
}

bool DoesNameExistInRegistry( const char * name)
{
    string command;
    command = "grep -i ";
    command += name;
    command += " /mnt/tmp/Windows/System32/config/SYSTEM";
    int ret = system (command.c_str());
    if (ret == 0) {
        cout << "Found " << name << " in the registry => Image is not clean\n" ;
        return true;
    }
    return false;
}

bool DoesDirectoryExists(const char * name)
{
    string command;
    command = "ls ";
    command += name;
    command += "> /dev/NULL";
    int ret = system (command.c_str());
    if (ret == 0) {
        return true;
    }
    return false;
}

const int LINE_SIZE =1024;

bool DoesFileExist(const char *dir, const char * pattern)
{
    char Line[LINE_SIZE];
    int count = 0;
    string s1;
    FILE *pipein_fp;
    s1 = "find ";
    s1 += dir ;
    s1 += " -iname ";
    s1 += pattern;

    cout << "Running command: " << s1 << "\n";

    pipein_fp = popen(s1.c_str(), "r");
    if(pipein_fp == NULL) {
        cerr << "find failed, assuming file exists\n";
        return true;
    }
    while(fgets(Line, LINE_SIZE, pipein_fp)) {
        count++;
        cerr << "find has found a match: " << Line << "\n";
    }
    
    if(count == 0) {
        return false;
    }
    return true;
}

bool VerifyCleanImage(vector <OS> v1, int ImageNumber)
{
    bool success;
    bool PartitionIsClean = false;
    string mount = "mount /dev/" +  v1[ImageNumber].Location;
    mount += " /mnt/tmp";
    bool FileExists = false;
    success = ExecuteCommand(mount);
    if(!success) {
        cerr << mount + " failed, assuming image is not clean\n";
        return false;
    }

    FileExists = DoesFileExist("/mnt/tmp/Windows/System32/drivers", "mlx4*");
    if (FileExists) {
        PartitionIsClean = false;
        goto Cleanup;
    }

    FileExists = DoesFileExist("/mnt/tmp/Windows/System32/drivers", "mthca*");
    if (FileExists) {
        PartitionIsClean = false;
        goto Cleanup;
    }

    FileExists = DoesFileExist("/mnt/tmp/Program\\ Files", "Mellanox");
    if (FileExists) {
        PartitionIsClean = false;
        goto Cleanup;
    }

// startup menu (2003 and 2008 are different)

    if (DoesDirectoryExists("/mnt/tmp/Documents\\ and\\ Settings/All\\ Users/Start\\ Menu/Programs/")) {
        FileExists = DoesFileExist("/mnt/tmp/Documents\\ and\\ Settings/All\\ Users/Start\\ Menu/Programs/", "Mellanox");
        if (FileExists) {
            PartitionIsClean = false;
            goto Cleanup;
        }
    }

    if (DoesDirectoryExists("/mnt/tmp/ProgramData/Microsoft/Windows/Start\\ Menu/Programs/")) {
        FileExists = DoesFileExist("/mnt/tmp/ProgramData/Microsoft/Windows/Start\\ Menu/Programs/", "Mellanox");
        if (FileExists) {
            PartitionIsClean = false;
            goto Cleanup;
        }
    }

    FileExists = DoesNameExistInRegistry("Mellanox");
    if (FileExists) {
        PartitionIsClean = false;
        goto Cleanup;
    }



    PartitionIsClean = true;
Cleanup:    

    string umount =  "umount /dev/" +  v1[ImageNumber].Location;
    success = ExecuteCommand(umount);
    if(!success) {
        cerr << umount + " failed\n";
        PartitionIsClean = false;
    }
    return PartitionIsClean;

}

void CreateImage(vector <OS> v1, int ImageNumber)
{
    bool ImageClean = VerifyCleanImage( v1, ImageNumber);
    if (!ImageClean) {
        cout << "Image is not clean => you can not back it up\n ";
        return;
    }

    string Image = ImageName(v1, ImageNumber);
    string command = "ntfsclone --save-image -o - /dev/";
    command += v1[ImageNumber].Location;
    command += " | gzip -c > ";
    command += Image;
    ExecuteCommand(command);
}

void RestoreImage(vector <OS> v1, int ImageNumber)
{
    string Image = ImageName(v1, ImageNumber);
    string command = "gunzip -c ";
    command += Image;
    command += " | ntfsclone --restore-image --overwrite /dev/";
    command += v1[ImageNumber].Location;
    command += " - ";    
    ExecuteCommand(command);

}



int  main(int argc, char **argv) 
{
    int Operation, OsNumber;
    bool IsRoot = false;
    vector <OS> v1 = ReadFile();
    PrintOSs(v1);
    cout << "Press 1 to create an image (folowed by image number)\n";
    cout << "Press 2 to restore an image (folowed by image number)\n";
    cout << "Press 3 to create all images (folowed by 0)\n";

    if ((argc == 2) && (!strcmp(argv[1],"/root"))) {
        IsRoot = true;
    }

    if ((argc == 3) && (!strcmp(argv[1],"/RestoreImage"))) {
        OsNumber = atoi(argv[2]);
        cout << "Restoring image " << OsNumber << "\n";
        RestoreImage(v1, OsNumber);
        return 0;
    }


    cin >> Operation >> OsNumber;
    cout << "Operation is " << Operation << " Os number is " << OsNumber << "\n";

    switch (Operation) {
        case 1:
            if (!IsRoot) {
                cout << "must be in root mode to create image,\n use " << argv[0]<< " /root to be in root mode \n";
                exit(3);                
            }
            CreateImage(v1, OsNumber); 
            break;

        case 2: 
            RestoreImage(v1, OsNumber); 
            break;

        case 3:
            {
                if (!IsRoot) {
                    cout << "must be in root mode to create image,\nuse " << argv[0]<< " /root to be in root mode \n";
                    exit(3);                
                }                
                for (int i=0; i < v1.size(); i++) {
                    CreateImage(v1, i); 
                }

            }
            break;
        default:
            cout << "Error Unknown option \n";
            exit(2);
    }


    return 0;
}

