// Measure the time that it takes to perform a context switch.

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

HANDLE *g_pEvents;

struct ThreadContext {
    int MyId;
    int NextId;
    int Count;
};

DWORD WINAPI ThreadFunc( LPVOID lpParam ) 
{ 
    ThreadContext *pContext = (ThreadContext*) lpParam;

    printf( "Thread %d starting, signaling thread %d\n", pContext->MyId, pContext->NextId);
    int j = pContext->Count;
    while (j-- != 0) {
        // Wait for my event
        WaitForSingleObject(g_pEvents[pContext->MyId], INFINITE);
        // Free the next thread
        SetEvent(g_pEvents[pContext->NextId]);
    }
    printf( "Thread %d exiting (not) freeing (%d)\n", pContext->MyId, pContext->NextId);
    return 0; 
} 

/*
    Test the time that it takes to do a context switch
*/

void RunTest(int NumberOfThreads, int CurrentlyRunning, int Count)
{
    int i;
    HANDLE *pThreads;
    ThreadContext *pContext;
    assert(NumberOfThreads % CurrentlyRunning == 0); 
    // If the assert above is not true, the test is not functioning 
    // in it's intended way.

    // Create a global list of events
    g_pEvents = new HANDLE [NumberOfThreads];    
    assert(g_pEvents != NULL);
    pThreads = new HANDLE [NumberOfThreads];    
    assert(pThreads != NULL);

    for (i = 0; i < NumberOfThreads; i++) {
        g_pEvents[i] = CreateEvent(NULL, FALSE, FALSE, NULL);
        assert(g_pEvents[i] != NULL);

        // Create the thread
        pContext = new ThreadContext;
        pContext->MyId = i;
        pContext->Count = Count;
        pContext->NextId = (i + CurrentlyRunning) % NumberOfThreads;
        
        pThreads[i] = CreateThread(NULL, 0 ,ThreadFunc ,pContext, 0,NULL );
        assert(pThreads[i] != NULL);
    }
    Sleep(2000);
    printf("Starting Threads\n");
    double Start = GetTickCount();
    for (i = 0 ; i < CurrentlyRunning; i++) {
        SetEvent(g_pEvents[i]);
    }

    // Wait for threads to die
    WaitForMultipleObjects(NumberOfThreads,pThreads, TRUE, INFINITE);
    double elapsed = (GetTickCount() - Start) / 1000;
    printf("Test Finished, %d context switches, time = %f"
        ", %f us to context swithch ", NumberOfThreads * Count, elapsed,elapsed / NumberOfThreads / Count * 1000000  );

}
void foo(int i)
{

}

// Test the time that it takes to enter & leave a critical section

void TestCriticalSection()
{
    int Count = 10000000, i;
    CRITICAL_SECTION CriticalSection;
    InitializeCriticalSection(&CriticalSection);
    HANDLE hEvent;

    hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);



    double Start = GetTickCount();
    for (i = 0 ; i < Count; i++) {
//        foo(i);

        // Test a CS
//        EnterCriticalSection(&CriticalSection);
//        LeaveCriticalSection(&CriticalSection);

        // Test an event
        SetEvent(hEvent);
        WaitForSingleObject(hEvent, INFINITE);

    
    }
    double elapsed = (GetTickCount() - Start) / 1000;
    printf("Test Finished, time per operation = %fus\n", elapsed /Count * 1000000 );

    


}


int MemorySpeed()
{
    int i,j;
    int DataSize = 1024 * 1024 * 10;
    int NumberToRepeat = 1 * 1024;

    DataSize = DataSize / 10;
    NumberToRepeat = NumberToRepeat * 10;

    char * p1 = (char *) malloc (DataSize);
    char * p2 = (char*) malloc (DataSize);

    DWORD dwStart = GetTickCount();

    for (i = 0; i < NumberToRepeat; i++) {
        memcpy(p1,p2,DataSize);
    }
    printf("Copying took %f seconds\n", (GetTickCount() - dwStart) / 1000.0);
    free (p2);
    free (p1);
    return 0;

}

int _cdecl main(int argc, char *argv[]) {

    //RunTest(30, 1, 300000);
    MemorySpeed();
    TestCriticalSection();

}



