#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <assert.h>
#include <sys/time.h>
#include <unistd.h>

pthread_mutex_t *g_pEvents;

struct ThreadContext {
    int MyId;
    int NextId;
    int Count;
};

int GetTickCount()
{
    struct timeval tv;
    int ret = gettimeofday(&tv, NULL);
    assert(ret ==0);
    return tv.tv_sec * 1000 + tv.tv_usec / 1000;
}

void *ThreadFunc( void *lpParam ) 
{ 
    ThreadContext *pContext = (ThreadContext*) lpParam;

    printf( "Thread %d starting, signaling thread %d\n", pContext->MyId, pContext->NextId);
    int j = pContext->Count;
    while (j-- != 0) {
        // Wait for my event
        pthread_mutex_lock(&g_pEvents[pContext->MyId]);
        // Free the next thread
        pthread_mutex_unlock(&g_pEvents[pContext->NextId]);
    }
    printf( "Thread %d exiting (not) freeing (%d)\n", pContext->MyId, pContext->NextId);
} 

/*
    Test the time that it takes to do a context switch
*/

void RunTest(int NumberOfThreads, int CurrentlyRunning, int Count)
{
    int i;
    pthread_t *pThreads;
    ThreadContext *pContext;
    assert(NumberOfThreads % CurrentlyRunning == 0); 
    // If the assert above is not true, the test is not functioning 
    // in it's intended way.

    // Create a global list of events
    g_pEvents = new pthread_mutex_t [NumberOfThreads];    
    assert(g_pEvents != NULL);
    pThreads = new pthread_t [NumberOfThreads];    
    assert(pThreads != NULL);

    for (i = 0; i < NumberOfThreads; i++) {
//???        g_pEvents[i] = PTHREAD_MUTEX_INITIALIZER;
        pthread_mutex_lock( &g_pEvents[i] );

        // Create the thread
        pContext = new ThreadContext;
        pContext->MyId = i;
        pContext->Count = Count;
        pContext->NextId = (i + CurrentlyRunning) % NumberOfThreads;
        
         int iRet= pthread_create(&pThreads[i], NULL, ThreadFunc ,pContext);
            //CreateThread(NULL, 0 ,ThreadFunc ,pContext, 0,NULL );
        assert(iRet == 0);
    }
    sleep(2);
    printf("Starting Threads\n");
    double Start = GetTickCount();
    for (i = 0 ; i < CurrentlyRunning; i++) {
        pthread_mutex_unlock( &g_pEvents[i] );
    }

    // Wait for threads to die
   for(int j=0; j < NumberOfThreads; j++)
   {
      pthread_join( pThreads[j], NULL); 
   }

    double elapsed = (GetTickCount() - Start) / 1000;
    printf("Test Finished, %d context switches, time = %f"
        ", %f us to context swithch ", NumberOfThreads * Count, elapsed,elapsed / NumberOfThreads / Count * 1000000  );

}

int main(int argc, char *argv[]) {

    RunTest(40, 8, 300000);
}
